# -*- coding: utf-8 -*-
"""
This file implements a proxy for accessing container file systems.
"""
__all__ = ('FileProxy',)

from typing import List, Union, Iterator, Optional, overload
from typing_extensions import Literal
import os
import contextlib
import shlex
import logging
import tempfile
import subprocess

from docker.models.containers import Container as DockerContainer

from .shell import ShellProxy
from ..exceptions import ROSWireException

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class FileProxy:
    def __init__(self,
                 container_docker: DockerContainer,
                 shell: ShellProxy
                 ) -> None:
        self.__docker_id = container_docker.id
        self.__container_docker = container_docker
        self.__shell = shell

    def copy_from_host(self, path_host: str, path_container: str) -> None:
        """
        Copies a given file or directory tree from the host to the container.

        Parameters:
            fn_host: the file or directory tree that should be copied from
                the host.
            fn_container: the destination path on the container.

        Raises:
            FileNotFoundError: if no file or directory exists at the given path
                on the host.
            FileNotFoundError: if the parent directory of the container
                filepath does not exist.
            OSError: if the copy operation failed.
        """
        id_container: str = self.__docker_id
        if not os.path.exists(path_host):
            m = f"file [{path_host}] does not exist on host"
            raise FileNotFoundError(m)

        path_container_parent: str = os.path.dirname(path_container)
        if not os.path.isdir(path_container_parent):
            m = (f"directory [{path_container_parent}] "
                 "does not exist on container")
            raise FileNotFoundError(m)

        cmd: str = (f"docker cp {shlex.quote(path_host)} "
                    f"{id_container}:{shlex.quote(path_container)}")
        try:
            subprocess.check_call(cmd, shell=True)
        except subprocess.CalledProcessError:
            m = (f"failed to copy file [{path_host}] "
                 f"from host to container [{id_container}]: {path_container}")
            raise OSError(m)

    def copy_to_host(self, path_container: str, path_host: str) -> None:
        """
        Copies a given file or directory tree from the container to the host.

        Parameters:
            fn_container: the file that should be copied from the container.
            fn_host: the destination filepath on the host.

        Raises:
            FileNotFoundError: if no file or directory exists at the given path
                inside the container.
            FileNotFoundError: if the parent directory of the host filepath
                does not exist.
            OSError: if the copy operation failed.
        """
        id_container: str = self.__docker_id
        if not self.exists(path_container):
            m = (f"file [{path_container}] does not exist in container "
                 f"[{id_container}]")
            raise FileNotFoundError(m)

        path_host_parent: str = os.path.dirname(path_host)
        if not os.path.isdir(path_host_parent):
            m = (f"directory [{path_host_parent}] "
                 "does not exist on host machine")
            raise FileNotFoundError(m)

        cmd: str = (f"docker cp {id_container}:{shlex.quote(path_container)} "
                    f"{shlex.quote(path_host)}")
        try:
            subprocess.check_call(cmd, shell=True)
        except subprocess.CalledProcessError:
            m = (f"failed to copy file [{path_container}] "
                 f"from container [{id_container}] to host: {path_host}")
            raise OSError(m)

    def write(self, fn: str, contents: Union[str, bytes]) -> None:
        mode = 'wb' if isinstance(contents, bytes) else 'w'
        dir_parent = os.path.dirname(fn)
        if not self.isdir(dir_parent):
            m = f"parent directory does not exist: {dir_parent}"
            raise FileNotFoundError(m)

        # write to a temporary file on the host and copy to container
        _, fn_host = tempfile.mkstemp()
        try:
            with open(fn_host, mode) as f:
                f.write(contents)
            self.copy_from_host(fn_host, fn)
        finally:
            os.remove(fn_host)

    @overload
    def read(self, fn: str) -> str:
        ...

    @overload
    def read(self, fn: str, binary: Literal[True]) -> bytes:
        ...

    @overload
    def read(self, fn: str, binary: Literal[False]) -> str:
        ...

    def read(self, fn: str, binary: bool = False) -> Union[str, bytes]:
        """
        Reads the contents of a given file.

        Parameters:
            fn: path to the file.
            binary: if true, read the binary contents of the file;
                otherwise, treat the file as a text file.
        """
        mode = 'rb' if binary else 'r'
        if not self.exists(fn):
            raise FileNotFoundError(f"file not found: {fn}")
        if self.isdir(fn):
            raise IsADirectoryError(f"cannot read directory: {fn}")

        _, fn_host_temp = tempfile.mkstemp(suffix='.roswire')
        try:
            self.copy_to_host(fn, fn_host_temp)
            with open(fn_host_temp, mode) as f:
                return f.read()
        finally:
            os.remove(fn_host_temp)

    def exists(self, path: str) -> bool:
        """
        Determines whether a file or directory exists at the given path.
        """
        cmd = f'test -e {shlex.quote(path)}'
        code, output, duration = self.__shell.execute(cmd)
        return code == 0

    def isfile(self, path: str) -> bool:
        """
        Determines whether a regular file exists at a given path.
        """
        cmd = f'test -f {shlex.quote(path)}'
        code, output, duration = self.__shell.execute(cmd)
        return code == 0

    def isdir(self, path: str) -> bool:
        """
        Determines whether a directory exists at a given path.
        """
        cmd = f'test -d {shlex.quote(path)}'
        code, output, duration = self.__shell.execute(cmd)
        return code == 0

    def islink(self, path: str) -> bool:
        """
        Determines whether a symbolic link exists at a given path.
        """
        cmd = f'test -h {shlex.quote(path)}'
        code, output, duration = self.__shell.execute(cmd)
        return code == 0

    def find(self,
             path: str,
             filename: str
             ) -> List[str]:
        """
        Returns a list of files that match the provided filename
        in a given directory recursively.

        Parameters:
            path: absolute path to the directory.
            filename: the name of the file to match.

        Raises:
            OSError: if find operation is not successful.
        """
        cmd = f'find {shlex.quote(path)} -name {shlex.quote(filename)}'
        code, output, duration = self.__shell.execute(cmd)
        if code != 0:
            raise OSError(f"Find operation was unsuccessful on {path}")
        paths: List[str] = output.replace('\r', '').split('\n')
        return paths

    def listdir(self,
                d: str,
                *,
                absolute: bool = False
                ) -> List[str]:
        """
        Returns a list of the files belonging to a given directory.

        Parameters:
            d: absolute path to the directory.
            absolute: if True, returns a list of absolute paths; if False,
                returns a list of relative paths.

        Raises:
            OSError: if the given path isn't a directory.
            OSError: if the given path is not a file or directory.
        """
        cmd = f'ls -A -1 {shlex.quote(d)}'
        code, output, duration = self.__shell.execute(cmd)
        if code == 2:
            raise OSError(f"no such file or directory: {d}")
        if not self.isdir(d):
            raise OSError(f"Not a directory: {d}")
        paths: List[str] = output.replace('\r', '').split('\n')
        if absolute:
            paths = [os.path.join(d, p) for p in paths]
        return paths

    def mkdir(self, d: str) -> None:
        """
        Creates a directory at a given path.

        Raises:
            NotADirectoryError: if the parent directory isn't a directory.
            FileNotFoundError: if the parent directory doesn't exist.
            FileExistsError: if a file or directory already exist at the given
                path.
            ROSWireException: if an unexpected error occurred.
        """
        cmd = f"mkdir {shlex.quote(d)}"
        code, output, duration = self.__shell.execute(cmd)
        if code == 0:
            return

        if self.exists(d):
            raise FileExistsError(d)

        d_parent = os.path.dirname(d)
        if not self.exists(d_parent):
            raise FileNotFoundError(d_parent)
        elif not self.isdir(d_parent):
            raise NotADirectoryError(d_parent)
        else:
            raise ROSWireException("unexpected mkdir failure")

    def makedirs(self, d: str, exist_ok: bool = False) -> None:
        """
        Recursively creates a directory at a given path, creating any missing
        intermediate directories along the way.

        Parameters:
            d: the path to the directory.
            exist_ok: specifies whether or not an exception should be raised
                if the given directory already exists.

        Raises:
            FileExistsError: if either (a) `exist_ok=False` and a directory
                already exists at the given path, or (b) a file already exists
                at the given path.
            NotADirectoryError: if the parent directory isn't a directory.
            ROSWireException: if an unexpected error occurred.
        """
        d_parent = os.path.dirname(d)
        if self.isdir(d) and not exist_ok:
            m = f"directory already exists: {d}"
            raise FileExistsError(m)
        if self.isfile(d):
            m = f"file already exists at given path: {d}"
            raise FileExistsError(m)
        if self.exists(d_parent) and self.isfile(d_parent):
            m = f"parent directory is actually a file: {d_parent}"
            raise NotADirectoryError(m)

        cmd = f'mkdir -p {shlex.quote(d)}'
        code, output, duration = self.__shell.execute(cmd)
        if code != 0:
            raise ROSWireException("unexpected makedirs failure")

    def remove(self, fn: str) -> None:
        """
        Removes a given file.

        Note:
            Does not handle permissions errors.

        Raises:
            FileNotFoundError: if the given file does not exist.
            IsADirectoryError: if the given path is a directory.
            ROSWireException: an unexpected failure occurred.
        """
        cmd = f'rm {shlex.quote(fn)}'
        code, output, duration = self.__shell.execute(cmd)
        if code == 0:
            return

        if not self.exists(fn):
            raise FileNotFoundError(f"file not found: {fn}")
        elif self.isdir(fn):
            raise IsADirectoryError(f"cannot remove directory: {fn}")

        raise ROSWireException("unexpected remove failure")

    def rmdir(self, d: str) -> None:
        """
        Removes a given directory.

        Note:
            Does not handle permissions errors.

        Raises:
            FileNotFoundError: if the given directory does not exist.
            NotADirectoryError: if the given path is not a directory.
            OSError: if the given directory is not empty.
            ROSWireException: an unexpected failure occurred.
        """
        if self.isfile(d):
            raise NotADirectoryError(f"cannot remove file: {d}")
        if not self.isdir(d):
            raise FileNotFoundError(f"directory does not exist: {d}")

        cmd = f'rmdir {shlex.quote(d)}'
        code, output, duration = self.__shell.execute(cmd)
        if code == 0:
            return
        if code == 1 and 'Directory not empty' in output:
            raise OSError(f"directory not empty: {d}")

        raise ROSWireException("unexpected rmdir failure")

    def rmtree(self, d: str) -> None:
        """
        Removes a given directory tree.

        Note:
            Does not handle permissions errors.

        Raises:
            FileNotFoundError: if the given directory does not exist.
            NotADirectoryError: if the given path is not a directory.
            OSError: if the directory tree removal failed.
        """
        if self.isfile(d):
            raise NotADirectoryError(f"cannot remove file: {d}")
        if not self.isdir(d):
            raise FileNotFoundError(f"directory does not exist: {d}")

        cmd = f'rm -rf {shlex.quote(d)}'
        code, output, duration = self.__shell.execute(cmd)
        if code != 0:
            raise OSError(f"failed to remove directory tree: {d}")

    def mktemp(self,
               suffix: Optional[str] = None,
               prefix: Optional[str] = None,
               dirname: Optional[str] = None
               ) -> str:
        """Creates a temporary file.

        Parameters
        ----------
        suffix: str, optional
            an optional suffix for the filename.
        prefix: str, optional
            an optional prefix for the filename.
        dirname: str, optional
            if specified, the temporary file will be created in the given
            directory.

        Raises
        ------
        FileNotFoundError:
            if specified directory does not exist.
        OSError:
            if the temporary file could not be constructed.

        Returns
        -------
        str
            The absolute path of the temporary file.
        """
        template = shlex.quote(f"{prefix if prefix else 'tmp'}.XXXXXXXXXX")
        cmd_parts = ['mktemp', template]
        dirname = dirname if dirname else '/tmp'
        if suffix:
            cmd_parts += ['--suffix', shlex.quote(suffix)]
        cmd_parts += ['-p', shlex.quote(dirname)]
        if not self.isdir(dirname):
            m = f'directory does not exist: {dirname}'
            raise FileNotFoundError(m)
        cmd = ' '.join(cmd_parts)

        code, output, duration = self.__shell.execute(cmd)
        # TODO capture context
        if code != 0:
            raise OSError(f"failed to create temporary directory")

        return output

    @contextlib.contextmanager
    def tempfile(self,
                 suffix: Optional[str] = None,
                 prefix: Optional[str] = None,
                 dirname: Optional[str] = None
                 ) -> Iterator[str]:
        """Creates a temporary file within a context.

        Upon exiting the context, the temporary file will be destroyed.

        See Also
        --------
        mktemp: Uses the same arguments to create a temporary file.

        Yields
        ------
        str
            The absolute path of the temporary file.
        """
        fn = self.mktemp(suffix=suffix, prefix=prefix, dirname=dirname)
        logger.debug("created temporary file: %s", fn)
        yield fn
        logger.debug("destroying temporary file: %s", fn)
        try:
            self.remove(fn)
        except FileNotFoundError:
            logger.debug("temporary file already destroyed: %s", fn)



import os
from glob import glob
from setuptools import setup, find_packages


path = os.path.join(os.path.dirname(__file__), 'src/roswire/version.py')
with open(path, 'r') as f:
    exec(f.read())


setup(
    name='roswire',
    version=__version__,
    python_requires='>=3.6',
    description='An API for testing and mutating ROS systems.',
    author='Chris Timperley',
    author_email='christimperley@googlemail.com',
    url='https://github.com/ChrisTimperley/roswire',
    license='mit',
    install_requires=[
        'attrs>=17.2.0',
        'typing-extensions>=3.7.2',
        'ruamel.yaml>=0.15.89',
        'toposort~=1.5',
        'docker~=3.7.2',
        'psutil~=5.6.2',
        'pyyaml~=5.1'
    ],
    setup_requires=[
        'pytest-runner'
    ],
    tests_require=[
        'pytest'
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    keywords=['ros', 'docker', 'testing', 'mutation', 'bug', 'repair'],
    classifiers=[
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    test_suites='tests'
)



from typing import Dict, Union


def decode_header(b: bytes) -> Dict[str, Union[int, str]]:
    contents: Dict[str, Union[int, str]] = {}
    length_header = int.from_bytes(b[0:4], byteorder='little')
    offset = 4

    # decode each field
    while offset < length_header:
        length = int.from_bytes(b[offset:offset + 4], byteorder='little')
        offset += 4
        name, _, val = b[offset:offset + length].decode('utf-8').partition('=')
        contents[name] = val
        offset += length

    return contents



# -*- coding: utf-8 -*-
# http://wiki.ros.org/Bags/Format/2.0
__all__ = ('BagRecorderProxy', 'BagPlayerProxy')

from typing import Optional, Collection
import logging
import shutil
import time
import pathlib
import threading
import subprocess
import os

from .file import FileProxy
from .shell import ShellProxy, Popen
from .node import NodeManagerProxy
from .. import exceptions

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class BagPlayerProxy:
    def __init__(self,
                 fn_container: str,
                 shell: ShellProxy,
                 files: FileProxy,
                 *,
                 delete_file_after_use: bool = False
                 ) -> None:
        self.__lock = threading.Lock()
        self.__fn_container = fn_container
        self.__shell = shell
        self.__files = files
        self.__delete_file_after_use = delete_file_after_use
        self.__started = False
        self.__stopped = False
        self._process: Optional[Popen] = None

    @property
    def started(self) -> bool:
        """Indicates whether or not playback has started."""
        return self.__started

    @property
    def stopped(self) -> bool:
        """Indicates whether or not playback has stopped."""
        return self.__stopped

    def __enter__(self) -> 'BagPlayerProxy':
        self.start()
        return self

    def __exit__(self, ex_type, ex_val, ex_tb) -> None:
        if ex_type is not None:
            logger.error("error occurred during bag playback",
                         exc_info=(ex_type, ex_val, ex_tb))
        if not self.stopped:
            self.stop()

    def finished(self) -> bool:
        """Checks whether playback has completed."""
        p = self._process
        return p.finished if p else False

    def wait(self, time_limit: Optional[float] = None) -> None:
        """Blocks until playback has finished.

        Parameters
        ----------
        time_limit: Optional[float] = None
            an optional time limit.

        Raises
        ------
        PlayerTimeout:
            if playback did not finish within the provided timeout.
        PlayerFailure:
            if an unexpected occurred during playback.
        """
        assert self._process
        try:
            self._process.wait(time_limit)
            retcode = self._process.returncode
            assert retcode is not None
            if retcode != 0:
                out = '\n'.join(self._process.stream)
                raise exceptions.PlayerFailure(retcode, out)
        except subprocess.TimeoutExpired:
            raise exceptions.PlayerTimeout

    def start(self) -> None:
        """Starts playback from the bag.

        Raises
        ------
        PlayerAlreadyStarted:
            if the player has already started.
        """
        logger.debug("starting bag playback")
        with self.__lock:
            if self.__started:
                raise exceptions.PlayerAlreadyStarted
            self.__started = True
            cmd: str = f"rosbag play -q {self.__fn_container}"
            self._process = self.__shell.popen(cmd)
            logger.debug("started bag playback")

    def stop(self) -> None:
        """Stops playback from the bag.

        Raises:
            PlayerAlreadyStopped: if the player has already been stopped.
        """
        logger.debug("stopping bag playback")
        with self.__lock:
            if self.__stopped:
                raise exceptions.PlayerAlreadyStopped
            if not self.__started:
                raise exceptions.PlayerNotStarted
            assert self._process
            self._process.kill()
            out = '\n'.join(list(self._process.stream))
            logger.debug("player output:\n%s", out)
            self._process = None
            if self.__delete_file_after_use:
                self.__files.remove(self.__fn_container)
            self.__stopped = True
        logger.debug("stopped bag playback")


class BagRecorderProxy:
    def __init__(self,
                 fn_dest: str,
                 ws_host: str,
                 shell: ShellProxy,
                 nodes: NodeManagerProxy,
                 exclude_topics: Optional[Collection[str]] = None
                 ) -> None:
        """
        Notes
        -----
            This object should not be constructed directly.

        Parameters
        ----------
        fn_dest: str
            the destination filepath for the bag (on the host).
        ws_host: str
            the workspace directory for the associated container (on the host).
        shell: ShellProxy
            a shell proxy.
        nodes: NodeManagerProxy
            access to nodes for the associated ROS graph.
        excluded_topics: Optional[Collection[str]] = None
            an optional list of topics that should be excluded from the bag.
        """
        self.__lock: threading.Lock = threading.Lock()
        self.__started: bool = False
        self.__stopped: bool = False
        self.__shell: ShellProxy = shell
        self.__nodes: NodeManagerProxy = nodes

        # FIXME generate a bag name
        self.__bag_name: str = "my_bag"

        # create a temporary file inside the shared directory
        self.__fn_host_dest: str = fn_dest
        self.__fn_container: str = f"/.roswire/{self.__bag_name}.bag"
        self.__fn_host_temp: str = \
            os.path.join(ws_host, f'{self.__bag_name}.bag')

    @property
    def started(self) -> bool:
        """Indicates whether or not recording has started."""
        return self.__started

    @property
    def stopped(self) -> bool:
        """Indicates whether or not recording has stopped."""
        return self.__stopped

    def __enter__(self) -> 'BagRecorderProxy':
        self.start()
        return self

    def __exit__(self, ex_type, ex_val, ex_tb) -> None:
        if ex_type is not None:
            logger.error("error occurred during bag recording",
                         exc_info=(ex_type, ex_val, ex_tb))
        should_save = ex_type is None
        if not self.stopped:
            self.stop(save=should_save)

    def start(self) -> None:
        """Starts recording to the bag.

        Raises
        ------
        RecorderAlreadyStarted:
            if the recorder has already been started.
        """
        logger.debug("starting bag recording")
        with self.__lock:
            if self.__started:
                raise exceptions.RecorderAlreadyStarted
            self.__started = True
            # FIXME bad mounting?!
            cmd: str = ("rosbag record -q -a"
                        f" -O {self.__fn_host_temp}"
                        f" __name:={self.__bag_name}")
            self.__shell.non_blocking_execute(cmd)
            logger.debug("started bag recording")

    def stop(self, save: bool = True) -> None:
        """Stops recording to the bag.

        Parameters
        ----------
        save: bool
            specifies whether the bag file should be saved to disk.

        Raises
        ------
        RecorderAlreadyStopped:
            if this recorder has already been stopped.
        """
        logger.debug("stopping bag recording")
        with self.__lock:
            if self.__stopped:
                raise exceptions.RecorderAlreadyStopped
            if not self.__started:
                raise exceptions.RecorderNotStarted

            name_node = f'/{self.__bag_name}'
            del self.__nodes[name_node]
            time.sleep(10)  # FIXME

            if os.path.exists(self.__fn_host_temp):
                if save:
                    shutil.copyfile(self.__fn_host_temp, self.__fn_host_dest)
                    logger.debug("bag file saved to %s", self.__fn_host_dest)
                else:
                    logger.debug("bag file will not be saved")
                os.remove(self.__fn_host_temp)
            else:
                logger.debug("temporary bag file not found on host: %s",
                             self.__fn_host_temp)
        logger.debug("stopped bag recording")



from roswire.proxy.launch import LaunchFileReader

from test_basic import build_file_and_shell_proxy


def test_read():
    with build_file_and_shell_proxy() as (shell, files):
        reader = LaunchFileReader(files, shell)
        reader.read('/ros_ws/src/mavros/mavros/launch/apm.launch',
                    argv=['foo:=bar', 'snoot:=boop'])
        assert False



# -*- coding: utf-8 -*-
__all__ = ('ContainerProxy', 'ContainerProxyManager')

from typing import Iterator, Optional, Union
from uuid import UUID, uuid4
from ipaddress import IPv4Address, IPv6Address
import ipaddress
import shlex
import contextlib
import logging
import shutil
import os

from docker import DockerClient
from docker import APIClient as DockerAPIClient
from docker.models.images import Image as DockerImage
from docker.models.containers import Container as DockerContainer

from .file import FileProxy
from .shell import ShellProxy
from ..util import Stopwatch
from ..exceptions import ROSWireException

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class ContainerProxy:
    def __init__(self,
                 api_docker: DockerAPIClient,
                 container_docker: DockerContainer,
                 uuid: UUID,
                 ws_host: str
                 ) -> None:
        self.__uuid = uuid
        self.__api_docker = api_docker
        self.__container_docker = container_docker
        self.__ws_host = ws_host

        info = api_docker.inspect_container(container_docker.id)
        self.__pid = int(info['State']['Pid'])
        self.__shell = ShellProxy(self.__api_docker,
                                  self.__container_docker,
                                  self.__pid)
        self.__files = FileProxy(self.__container_docker, self.__shell)

    @property
    def uuid(self) -> UUID:
        return self.__uuid

    @property
    def pid(self) -> int:
        """The PID of this container process on the host machine."""
        return self.__pid

    @property
    def shell(self) -> ShellProxy:
        return self.__shell

    @property
    def files(self) -> FileProxy:
        return self.__files

    @property
    def ws_host(self) -> str:
        return self.__ws_host

    @property
    def ip_address(self) -> str:
        container = self.__container_docker
        api = self.__api_docker
        docker_info = api.inspect_container(container.id)
        address = docker_info['NetworkSettings']['IPAddress']
        try:
            return str(IPv4Address(address))
        except ipaddress.AddressValueError:
            return str(IPv6Address(address))

    def persist(self,
                repo: Optional[str] = None,
                tag: Optional[str] = None
                ) -> DockerImage:
        """Persists this container to an image.

        Parameters
        ----------
        repo: str, optional
            The name of the repository to which the image should belong.
        tag: str, optional
            The tag that should be used for the image.

        Returns
        -------
        A description of the persisted image.
        """
        return self.__container_docker.commit(repo, tag)


class ContainerProxyManager:
    def __init__(self,
                 client_docker: DockerClient,
                 api_docker: DockerAPIClient,
                 dir_host_ws: str
                 ) -> None:
        self.__dir_host_ws = dir_host_ws
        self.__client_docker = client_docker
        self.__api_docker = api_docker

    @property
    def client_docker(self) -> DockerClient:
        return self.__client_docker

    @property
    def api_docker(self) -> DockerAPIClient:
        return self.__api_docker

    @contextlib.contextmanager
    def _build_shared_directory(self, uuid: UUID) -> Iterator[str]:
        dir_host = os.path.join(self.__dir_host_ws, 'containers', uuid.hex)
        try:
            logger.debug("creating container directory: %s", dir_host)
            os.makedirs(dir_host, exist_ok=True)
            logger.debug("created container directory: %s", dir_host)
            yield dir_host
        finally:
            if os.path.exists(dir_host):
                logger.debug("destroying container directory: %s", dir_host)
                shutil.rmtree(dir_host)
                logger.debug("destroyed container directory: %s", dir_host)

    @contextlib.contextmanager
    def _container(self,
                   image_or_name: Union[str, DockerImage],
                   dir_host_shared: str,
                   uuid: UUID
                   ) -> Iterator[DockerContainer]:
        cmd_env_file = ("export | tee /.environment > /dev/null && "
                        "chmod 444 /.environment && "
                        "echo 'ENVFILE CREATED' && /bin/bash")
        cmd_env_file = f"/bin/bash -c {shlex.quote(cmd_env_file)}"
        logger.debug("building docker container: %s", uuid)
        dockerc = self.__client_docker.containers.create(
            image_or_name,
            cmd_env_file,
            user='root',
            name=uuid,
            volumes={dir_host_shared: {'bind': '/.roswire', 'mode': 'rw'}},
            stdin_open=True,
            tty=False,
            detach=True)
        logger.debug("created docker container")
        try:
            dockerc.start()
            logger.debug("started docker container")

            # wait until .environment file is ready
            env_is_ready = False
            for line in self.__api_docker.logs(dockerc.id, stream=True):
                line = line.strip().decode('utf-8')
                if line == 'ENVFILE CREATED':
                    env_is_ready = True
                    break
            assert env_is_ready

            logger.debug("finished building container: %s", uuid)
            yield dockerc
        finally:
            logger.debug("destroying docker container: %s", uuid)
            dockerc.remove(force=True)
            logger.debug("destroyed docker container: %s", uuid)

    @contextlib.contextmanager
    def launch(self,
               image_or_name: Union[str, DockerImage]
               ) -> Iterator['ContainerProxy']:
        api_docker = self.__api_docker
        uuid = uuid4()
        logger.debug("UUID for container: %s", uuid)
        with self._build_shared_directory(uuid) as dir_shared:
            with self._container(image_or_name, dir_shared, uuid) as dockerc:
                c = ContainerProxy(api_docker, dockerc, uuid, dir_shared)
                yield c

    def image(self, tag: str) -> DockerImage:
        """Retrieves the Docker image with a given tag."""
        return self.__client_docker.images.get(tag)

    def image_sha256(self, tag_or_image: Union[str, DockerImage]) -> str:
        """Retrieves the SHA256 for a given Docker image."""
        image: DockerImage
        if isinstance(tag_or_image, str):
            image = self.image(tag_or_image)
        else:
            image = tag_or_image
        return image.id[7:]
